import PdfView from './index.tsx'
import MarkFormField from '../MarkFormField'
import { CurrentUserMark, Mark } from '../../types/mark.ts'
import React, { useState, useEffect } from 'react'
import {
  findNextIncompleteCurrentUserMark,
  getUpdatedMark,
  updateCurrentUserMarks
} from '../../utils'
import { EMPTY } from '../../utils/const.ts'
import { Container } from '../Container'
import signPageStyles from '../../pages/sign/style.module.scss'
import { CurrentUserFile } from '../../types/file.ts'
import FileList from '../FileList'
import { StickySideColumns } from '../../layouts/StickySideColumns.tsx'
import { UsersDetails } from '../UsersDetails.tsx'
import { Meta } from '../../types'
import {
  faCircleInfo,
  faFileDownload,
  faPen
} from '@fortawesome/free-solid-svg-icons'
import { Typography } from '@mui/material'
import styles from '../UsersDetails.tsx/style.module.scss'

interface PdfMarkingProps {
  currentUserMarks: CurrentUserMark[]
  files: CurrentUserFile[]
  /**
   * Currently, loading spinner is present if `files` array is of length 0,
   * Which means if no files are found, loading spinner will be spinning indefinitely
   * For that reason `noFiles` is introduced to set the loading off when fetching is finished.
   */
  noFiles?: boolean
  handleExport: () => void
  handleEncryptedExport: () => void
  handleSign: () => void
  meta: Meta | null
  otherUserMarks: Mark[]
  setCurrentUserMarks: (currentUserMarks: CurrentUserMark[]) => void
  setUpdatedMarks: (markToUpdate: Mark) => void
}

/**
 * Top-level component responsible for displaying Pdfs, Pages, and Marks,
 * as well as tracking if the document is ready to be signed.
 * @param props
 * @constructor
 */
const PdfMarking = (props: PdfMarkingProps) => {
  const {
    files,
    noFiles,
    currentUserMarks,
    setCurrentUserMarks,
    setUpdatedMarks,
    handleExport,
    handleEncryptedExport,
    handleSign,
    meta,
    otherUserMarks
  } = props
  const [selectedMark, setSelectedMark] = useState<CurrentUserMark | null>(null)
  const [selectedMarkValue, setSelectedMarkValue] = useState<string>('')
  const [currentFile, setCurrentFile] = useState<CurrentUserFile | null>(null)

  useEffect(() => {
    if (selectedMark === null && currentUserMarks.length > 0) {
      setSelectedMark(
        findNextIncompleteCurrentUserMark(currentUserMarks) ||
          currentUserMarks[0]
      )
    }
  }, [currentUserMarks, selectedMark])

  useEffect(() => {
    if (currentFile === null && files.length > 0) {
      setCurrentFile(files[0])
    }
  }, [files, currentFile])

  const handleMarkClick = (id: number) => {
    const nextMark = currentUserMarks.find((mark) => mark.mark.id === id)

    if (nextMark) handleCurrentUserMarkChange(nextMark)
  }

  const handleCurrentUserMarkChange = (mark: CurrentUserMark) => {
    if (!selectedMark) return
    const updatedSelectedMark: CurrentUserMark = getUpdatedMark(
      selectedMark,
      selectedMarkValue
    )

    const updatedCurrentUserMarks = updateCurrentUserMarks(
      currentUserMarks,
      updatedSelectedMark
    )
    setCurrentUserMarks(updatedCurrentUserMarks)

    // If clicking on the same mark, don't update the value, otherwise do update
    if (mark.id !== selectedMark.id) {
      setSelectedMarkValue(mark.currentValue ?? EMPTY)
      setSelectedMark(mark)
    }
  }

  /**
   * Sign and Complete
   */
  const handleSubmit = (event: React.MouseEvent<HTMLButtonElement>) => {
    event.preventDefault()
    if (selectedMarkValue && selectedMark) {
      const updatedMark: CurrentUserMark = getUpdatedMark(
        selectedMark,
        selectedMarkValue
      )

      setSelectedMarkValue(EMPTY)
      const updatedCurrentUserMarks = updateCurrentUserMarks(
        currentUserMarks,
        updatedMark
      )
      setCurrentUserMarks(updatedCurrentUserMarks)
      setSelectedMark(null)
      setUpdatedMarks(updatedMark.mark)
    }

    handleSign()
  }

  // const updateCurrentUserMarkValues = () => {
  //   const updatedMark: CurrentUserMark = getUpdatedMark(selectedMark!, selectedMarkValue)
  //   const updatedCurrentUserMarks = updateCurrentUserMarks(currentUserMarks, updatedMark)
  //   setSelectedMarkValue(EMPTY)
  //   setCurrentUserMarks(updatedCurrentUserMarks)
  // }

  const handleChange = (value: string) => {
    setSelectedMarkValue(value)
  }

  const renderRightColumn = () => {
    if (meta !== null) return <UsersDetails meta={meta} />

    return (
      <div className={styles.container}>
        <div className={styles.section}>
          <Typography>No meta found</Typography>
        </div>
      </div>
    )
  }

  return (
    <>
      <Container className={signPageStyles.container}>
        <StickySideColumns
          left={
            <div>
              {currentFile !== null && (
                <FileList
                  files={files}
                  currentFile={currentFile}
                  setCurrentFile={setCurrentFile}
                  handleExport={handleExport}
                  handleEncryptedExport={handleEncryptedExport}
                />
              )}
            </div>
          }
          right={renderRightColumn()}
          leftIcon={faFileDownload}
          centerIcon={faPen}
          rightIcon={faCircleInfo}
        >
          <PdfView
            currentFile={currentFile}
            files={files}
            noFiles={noFiles}
            handleMarkClick={handleMarkClick}
            selectedMarkValue={selectedMarkValue}
            selectedMark={selectedMark}
            currentUserMarks={currentUserMarks}
            otherUserMarks={otherUserMarks}
          />
          {noFiles && (
            <Typography textAlign="center">
              We were not able to retrieve the files.
            </Typography>
          )}
        </StickySideColumns>

        {!noFiles && (
          <MarkFormField
            handleSubmit={handleSubmit}
            handleSelectedMarkValueChange={handleChange}
            selectedMark={selectedMark}
            selectedMarkValue={selectedMarkValue}
            currentUserMarks={currentUserMarks}
            handleCurrentUserMarkChange={handleCurrentUserMarkChange}
          />
        )}
      </Container>
    </>
  )
}

export default PdfMarking
